#include <utility>
#include <cstdio>
#include <vector>
#include <queue>
#include <map>
#include <fcntl.h>

#include "buffer.hh"

template<typename T>
class ringbuf
{
    typedef std::vector<T> vec;
    std::queue<vec> buffers;
    unsigned frontptr;
public:
    ringbuf() : frontptr(0)
    {
    }
    
    void PushBack(const T* buf, unsigned n)
    {
        buffers.push(vec(buf, buf+n));
    }
    void PopFront(unsigned n)
    {
        while(n > 0 && !buffers.empty())
        {
            unsigned c = GetFrontSize();
            if(n < c)
            {
                frontptr += n;
                c = n;
            }
            else
            {
                buffers.pop();
                frontptr = 0;
            }
            n -= c;
        }
    }
    unsigned GetFrontSize() const
    {
        if(buffers.empty()) return 0;
        return buffers.front().size() - frontptr;
    }
    const unsigned char *GetFrontPtr() const
    {
        return &buffers.front().front() + frontptr;
    }
};

typedef ringbuf<unsigned char> buffer;

static std::map<int, buffer> buffers;

void AddWriteTask(FILE *fp, const unsigned char *data, unsigned size)
{
    MoveBuffers();
    
    int fd = fileno(fp);
    
    std::map<int, buffer>::iterator i;
    
    i = buffers.lower_bound(fd);
    if(i == buffers.end() || i->first != fd)
    {
        i = buffers.insert(i, std::make_pair(fd, buffer()));
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    }
    
    buffer& buf = i->second;
    buf.PushBack(data, size);
}

void MoveBuffers()
{
    std::map<int, buffer>::iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        int fd = i->first;
        buffer& buf = i->second;
        
        unsigned count = buf.GetFrontSize();
        const unsigned char *p = buf.GetFrontPtr();
        
        int n = write(fd, p, count);
        if(n > 0)
        {
            buf.PopFront(n);
        }
    }
}

void FlushBuffers()
{
    std::map<int, buffer>::iterator i,j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        int fd = i->first;

        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) &~ O_NONBLOCK);

        buffer& buf = i->second;
        
        for(;;)
        {
            unsigned count = buf.GetFrontSize();
            if(!count) break;
            
            const unsigned char *p = buf.GetFrontPtr();
            
            int n = write(fd, p, count);
            if(n > 0)
            {
                buf.PopFront(n);
            }
        }
        buffers.erase(i);
    }
}
