#include "settings.hh"
#include "x.hh"

#include <unistd.h> /* usleep */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xmu/Xmu.h>

#ifdef USE_SHM
#include <sys/ipc.h>
#include <sys/shm.h>   
#include <X11/extensions/XShm.h>
#endif

unsigned width, height, depth;

static Display *mDisplay;
static int mScreen;
static Window mRootWin, window;
static XVisualInfo vinfo;

static XImage *image;

#ifdef USE_SHM
static XShmSegmentInfo Shminfo;
#endif

static int Error(Display *dpy, XErrorEvent *err)
{
  (void) XmuPrintDefaultErrorMessage (dpy, err, stderr);
  return 0;
}

static void Failed(const char *what)
{
    fprintf(stderr, "%s failed...\n", what);
    fflush(stderr);
}

static Window XWindowByName(Display *display, const Window rootwin, const char *name)
{
    unsigned int num_children;
    Window *children, child, window;
    XTextProperty windowname;

    if(XGetWMName(display, rootwin, &windowname) != 0)
    {
        fprintf(stderr, "Window='%s'\n", (const char *)windowname.value);
        if(!strcmp((const char *)windowname.value, name))
            return rootwin;
    }
    
    window = (Window) NULL;
     
    if(XQueryTree(display, rootwin, &child, &child, &children, &num_children))
    {
        unsigned i;
        for(i=0; i < num_children; ++i)
        {
            /* Search each child and their children. */
            window = XWindowByName(display, children[i], name);
            if(window != (Window) NULL)
                break;
        }
        if (children != (Window *)NULL)
            XFree((void *)children);
    }
    return window;
}
static void XListWindows(Display *display, const Window rootwin)
{
    unsigned int num_children;
    Window *children, child;
    XTextProperty windowname;

    if(XGetWMName(display, rootwin, &windowname) != 0)
        printf("  '%s'\n", (const char *)windowname.value);
    
    if(XQueryTree(display, rootwin, &child, &child, &children, &num_children))
    {
        unsigned i;
        for(i=0; i < num_children; ++i)
        {
            /* Search each child and their children. */
            XListWindows(display, children[i]);
        }
        if (children != (Window *)NULL)
            XFree((void *)children);
    }
}

static void GetWindowGeometry()
{
    unsigned border_width;
    int xpos, ypos;
    Window root;
    
    fprintf(stderr, "XGetGeometry...\n");
    if(!XGetGeometry(
                 mDisplay, window, &root,
                 &xpos, &ypos,
                 &width, &height,
                 &border_width,
                 &depth))
    {
        Failed("XGetGeometry");
    }
    fprintf(stderr, "width=%u, height=%u, depth=%u\n", width, height, depth);
    
    width  -= XOFFSET;
    height -= YOFFSET;
}

void XInit()
{
    char *dispName = XDisplayName(NULL);
    unsigned num_tries;

    XSetErrorHandler(Error);

    mDisplay = XOpenDisplay(dispName);
    if(!mDisplay)
    {
        fprintf(stderr, "XOpenDisplay(%s) failed\n", dispName);
        return;
    }
    mScreen = DefaultScreen(mDisplay);
    
    mRootWin = RootWindow(mDisplay, mScreen);
    
    for(num_tries=0;;)
    {
        window = XWindowByName(mDisplay, mRootWin, WINDOW_NAME);
        if(window) break;
        if(++num_tries == 100)
        {
            Failed("XWindowByName");
            return;
        }
        usleep(20000);
    }
    
    XFlush(mDisplay);
    XSync(mDisplay, False);
#ifdef USE_SHM
    if(!XShmQueryExtension(mDisplay))
    {
        Failed("XShmQueryExtension");
        return;
    }
#endif
    GetWindowGeometry();

    if(!XMatchVisualInfo(mDisplay, mScreen, depth, DirectColor, &vinfo))
    {
        Failed("XMatchVisualInfo");
        return;
    }
}

unsigned char *GetImage()
{
#ifdef USE_SHM
    if(!image)
    {
        fprintf(stderr, "XShmCreateImage...\n");
        image = XShmCreateImage(mDisplay, vinfo.visual,
                  depth,
                  ZPixmap,
                  NULL,
                  &Shminfo,
                  width,
                  height);
        if(!image)
        {
            Failed("XShmCreateImage");
            return NULL;
        }
        
        fprintf(stderr, "shmget...\n");
        Shminfo.shmid = shmget(IPC_PRIVATE,
                               image->bytes_per_line * image->height,
                               IPC_CREAT | 0777);
        if(Shminfo.shmid < 0)
        {
            perror("shmget");
            XDestroyImage(image); image=NULL;
            return NULL;
        }
        fprintf(stderr, "shmat...\n");
        Shminfo.shmaddr = (char *) shmat(Shminfo.shmid, 0, 0);
        if(Shminfo.shmaddr == (char *)-1)
        {
            perror("shmat");
            XDestroyImage(image);
            return NULL;
        }
        fprintf(stderr, "shmctl...\n");
        shmctl(Shminfo.shmid, IPC_RMID, 0),

        image->data = Shminfo.shmaddr;
        Shminfo.readOnly = False;
        fprintf(stderr, "XShmAttach...\n");
        XShmAttach(mDisplay, &Shminfo);
        
        XSync(mDisplay, False);
    }
#endif
    /* If SHM failed or was disabled, try non-SHM way */
    if(!image)
    {
        /*fprintf(stderr, "XGetImage...\n");
        fflush(stderr);*/
        
        image = XGetImage(mDisplay, window,
                  XOFFSET, YOFFSET,
                  width,
                  height,
                  AllPlanes,
                  XYPixmap);
        if(!image)
        {
            Failed("XGetImage");
        }
    }
#ifdef USE_SHM
    /*fprintf(stderr, "XSHMGetImage...\n");
    fflush(stderr);*/
    if(!XShmGetImage(mDisplay, window, image,
                    XOFFSET, YOFFSET,
                    AllPlanes))
    {
        Failed("XSHMGetImage");
    }
#endif
    return image ? (unsigned char *)image->data : NULL;
}

void DeleteImage(void)
{
    if(!image) return;

#ifndef USE_SHM
    fprintf(stderr, "Destroying image\n");
    fflush(stderr);
    
    fprintf(stderr, "XDestroyImage...\n");
    XDestroyImage(image);
    image = NULL;
#endif
}

void XDone(void)
{
    if(!image) return;
    
#ifdef USE_SHM
    fprintf(stderr, "Destroying image\n");
    fflush(stderr);

    fprintf(stderr, "XShmDetach...\n");
    XShmDetach(mDisplay, &Shminfo);
    fprintf(stderr, "XDestroyImage...\n");
    XDestroyImage(image);
    image = NULL;
    fprintf(stderr, "shmdt...\n");
    shmdt(Shminfo.shmaddr);
#endif
}
